#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-12-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_median_blur (image, drawable, uradius):
    radius=float(uradius)
    percentile=50.0
    alpha_percentile=50.0
    gegl_graph_string="gegl:median-blur radius=%f percentile=%f alpha-percentile=%f" % (radius, percentile, alpha_percentile)
    gegl_graph(image, drawable, gegl_graph_string)

def gegl_gaussian_blur (image, drawable, ux, uy):    # gaussian blur  
    std_dev_x=float(ux)    #1.50     #user_range   (0.24, 100.0)
    std_dev_y=float(uy)    #1.50     #user_range   (0.24, 100.0)
    mystring="gegl:gaussian-blur std-dev-x=%f std-dev-y=%f"
    gegl_graph_string=mystring % (std_dev_x, std_dev_y)
    gegl_graph(image, drawable,gegl_graph_string)
    
def gegl_unsharp_mask (image, drawable, usdev , uscale, uthreshold):      
    std_dev=usdev           #user_range   (0.0, 40.0)
    scale=uscale            #user_range   (0.0, 10.0)
    threshold=uthreshold    #user_range   (0.0, 1.0) 
    mystring="gegl:unsharp-mask std-dev=%f scale=%f threshold=%f"
    gegl_graph_string=mystring % (std_dev, scale, threshold)
    gegl_graph(image, drawable,gegl_graph_string)

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

#---------------------------------------------------------------------------
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
    
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
    # adjusting values
    newBG = adjustmeanSV (newImage, newBG)
    
    # remove extremes
    pdb.gimp_drawable_curves_spline (newBG, HISTOGRAM_VALUE, 6,
                                     (0.00,0.05,
                                      0.50,0.50,
                                      1.00,0.95))
    newBG.opacity = 100.0
    
    return(newImage)
    
#---------------------------------------------------------------------------
# adjust mean values and mean saturation
#---------------------------------------------------------------------------
def adjustmeanSV (inImage, inLayer):
    meanSadj=0
    meanVadj=50
    # decompose HSV
    (HSVI,HL,SL,VL) = pdb.plug_in_decompose (inImage, inLayer, "HSV", True)
    #pdb.gimp_display_new (HSVI)
    #info_msg ("meanVadj="+str(meanVadj)+", meanSadj="+str(meanSadj))
    HL=HSVI.layers[0]; SL=HSVI.layers[1]; VL=HSVI.layers[2];
    if meanSadj!=0:
        #info_msg ("calling adjustmeanv for saturation")
        adjustmeanv (HSVI, SL, meanSadj)
    if meanVadj!=0:
        #info_msg ("calling adjustmeanv for value")
        adjustmeanv (HSVI, VL, meanVadj)
    pdb.plug_in_recompose (HSVI, HL)
    return(inLayer)

#---------------------------------------------------------------------------
# adjust mean values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer, meanadj):
    # take care of the existing mean in determining the theoretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    thmean = (127.5*meanadj/100+meanl*(100.0-meanadj)/100)   
    #info_msg ("thmean="+str(thmean))
    # range for the loops
    minmean = thmean*1.05  
    maxmean = thmean*1.15  
    
    modified=False
    if meanl < minmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        loop=0
        while meanl < minmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean and thmean!=meanl:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        loop=0
        while meanl > maxmean and loop<10:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
            loop+=1
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))

    return(inLayer)
    
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   
def JvidEffectX2b (inImage, inLayer, widthheight,   
                 blurval, 
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
        
    OrigWH = inLayer.width+inLayer.height
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # prepare image for effects Blur + USM 
    # - Blur  
    Blurred = Preprocessed.copy()
    newImage.add_layer (Blurred, 0)
    uRadius=float(blurval)                          # chg Gaussian to Median
    Blurred.name = "Blur applied "+str(blurval) 
    gegl_median_blur (newImage, Blurred, uRadius)
    # - USM
    uSdev=float(blurval);uScale=1.5; uThreshold=0.0       # chg uScale from 1.0
    Blurred.name = "Median Blur "+str(uSdev)+" then USM "+str(uSdev)  
    gegl_unsharp_mask (newImage, Blurred, uSdev, uScale, uThreshold)
    
    #pdb.gimp_display_new (newImage)
    #return

    # Effect: Lylejk_Luma_Invert 1.035,0    # chg to 2.005,0
    pdb.plug_in_gmic_qt(newImage, Blurred, 1,2,
    '-v - -Lylejk_Luma_Invert 2.005,0 ')
    Luma_Invert = newImage.layers[0]
    Luma_Invert.mode = LAYER_MODE_GRAIN_EXTRACT
    Luma_Invert.opacity = 100.0  
    Luma_Invert.name = '[GMIC] -Lylejk_Luma_Invert 2.005,0'
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Input for further step
    Base1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base1, 0)
    Base1.name = "Base 1 "
    Base1.mode = LAYER_MODE_HSV_VALUE
    pdb.gimp_item_set_visible (Luma_Invert, False)
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # Input for further step
    Base2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base2, 0)
    Base2.name = "Base for Isophotes"
    Base2.mode = LAYER_MODE_NORMAL
    Base2.opacity = 100.0
    
    # Effect: samj_Isophotes_Vers_Aquarelle
    #         16,0.5,8,0,1,10,1,0,0,0,0,1,10,10,4,3,251,237,206,1   chg 4 to 8
    Iso_Parms = '16,0.5,8,0,1,10,1,0,0,0,0,1,10,10,4,3,251,237,206,1'
    pdb.plug_in_gmic_qt(newImage, Base2, 1,2,
    '-v - -samj_Isophotes_Vers_Aquarelle '+Iso_Parms+' ')
    Iso = newImage.layers[0]
    Iso.mode = LAYER_MODE_OVERLAY   # chg from HSV_VALUE
    Iso.opacity = 70.0  
    Iso.name = '[GMIC] -samj_Isophotes_Vers_Aquarelle '+Iso_Parms
    ux=float(4.5); uy=ux
    gegl_gaussian_blur (newImage, Iso, ux, uy)
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Input for further step
    Base3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base3, 0)
    Base3.name = "Base 3 "
    Base3.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # source on top posterized     # chg G'mic Posterize from Gimp Posterize
    pdb.gimp_image_set_active_layer (newImage, Preprocessed)
    # Effect: fx_posterize 150,30,1,12,0,0,0,0,50,50    # chg 12 from 7
    posterize_Parms = '150,30,1,12,0,0,0,0,50,50'
    pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2,
    '-v - -fx_posterize '+posterize_Parms+' ')
    Poster = newImage.layers[0]
    Poster.name = " Preprocessed Poster "     # added
    Poster.mode = LAYER_MODE_GRAIN_MERGE
    Poster.opacity = 100.0
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # Input for further step
    Base4 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base4, 0)
    Base4.name = "Base for Morpho "
    Base4.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # Effect: fx_MorphoPaint 1,18,2,25,100,230,8,3,4,0.5,2,0.5,200,1,1,1,1,1,1,1,1,0
    #                 chg >  1,18,2,25,100,230,8,10,4,5,2,0.5,200,1,1,0.25,1,1,1,1,1,0
    MorphoPaint_Parms = '1,18,2,25,100,230,8,10,4,5,2,0.5,200,1,1,0.25,1,1,1,1,1,0'
    pdb.plug_in_gmic_qt(newImage, Base4, 1,2,
    '-v - -fx_MorphoPaint '+MorphoPaint_Parms+' ')
    MorphoPaint = newImage.layers[0]
    MorphoPaint.mode = LAYER_MODE_MULTIPLY
    MorphoPaint.opacity = 70.0
    MorphoPaint.name = '[GMIC] -fx_MorphoPaint '+MorphoPaint_Parms
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Input for further step
    Base5 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base5, 0)
    Base5.name = "Base for Vector Paint "
    Base5.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)    
    #return
    
    # Effect Vector Painting
    # [G'MIC]  fx_vector_painting 9.7,0,50,50
    pdb.plug_in_gmic_qt(newImage, Base5, 1,2 ,
    '-v - -fx_vector_painting 9.7,0,50,50 ')
    Vector_Paint = newImage.layers[0]
    Vector_Paint.opacity = 80.0  # back to v1
    Vector_Paint.mode = LAYER_MODE_NORMAL  # back to v1
    Vector_Paint.name = "[GMIC] -fx_vector_painting 9.7,0,50,50 "
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Input for further step (over canvas)
    Base6 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base6, 0)
    pdb.gimp_context_set_background ((128,120,112,255))
    pdb.gimp_drawable_fill (Base6, FILL_BACKGROUND)
    pdb.plug_in_gmic_qt(newImage, Base6, 1,0,       
    "-v - -texturize_canvas 20,4,0.8 ")
    Canvas = newImage.layers[0]
    Canvas.mode = LAYER_MODE_LINEAR_LIGHT    # chg from MULTIPLY
    Canvas.opacity = 95.0   # chg from 50.0   
    uSdev=3.0;uScale=2.5; uThreshold=0.0       # strong sharpen canvas texture  chg 2.5
    gegl_unsharp_mask (newImage, Canvas, uSdev, uScale, uThreshold)
    Canvas.name = "canvas 20,4,0.8 USM 3,2 "
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # addition
    # Input for final adjustments
    Base7 = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (Base7, 0)
    Base7.name = "Base for Final adjustments"
    uSdev=3.0;uScale=1.5; uThreshold=0.0       # strong sharpen outcome
    gegl_unsharp_mask (newImage, Base7, uSdev, uScale, uThreshold)
    Poster2 = Poster.copy()
    newImage.add_layer (Poster2, 0)
    Poster2.name = " Poster (dup)"     
    Poster2.mode = LAYER_MODE_GRAIN_MERGE
    Poster2.opacity = 40.0  ###### instead of 70
    PosterD = Poster2.copy()
    newImage.add_layer (PosterD, 0)
    PosterD.name = " Poster (dup) desat"     
    pdb.gimp_drawable_desaturate (PosterD, DESATURATE_LUMA)
    PosterD.mode = LAYER_MODE_NORMAL
    PosterD.opacity = 10.0  # chg from 20.0
    FinalAdj = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
    newImage.add_layer (FinalAdj, 0)
    FinalAdj.name = "Final adjustments: USM + Poster5"
    FinalAdj.mode = LAYER_MODE_NORMAL   # back to NORMAL from SCREEN
    pdb.gimp_drawable_curves_spline (FinalAdj, HISTOGRAM_VALUE, # remove some blacks&white
                                     4, (0.0,0.001,1.0,0.98))  
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        pdb.gimp_image_remove_layer (newImage, Luma_Invert)
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect X2b (Graphic_Effect_Illustration upd v.2)"
    outcome.name = outcome.name+", parms="
    outcome.name = outcome.name+str(blurval)
    
    if widthheight==0: widthheight=OrigWH
    uRatio = widthheight*1.00/FixWH*1.00
    uWidth = int(width*uRatio)
    uHeight = int(height*uRatio)
    #info_msg ("SCALING")   # not during tests, scaling is slow
    pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect X2b (Graphic_Effect_Illustration upd v.2) "

register(
    "JvidEffectX2b",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_ADJUSTMENT, "softglow", "Blur amount", 3.5, (1.0, 7.5, 0.1)),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectX2b,
    menu=menu
)

main()
